﻿/*
 * Mr Action Bar Populatorerer (Proof of Concept / Beta) - by Any (Anything)

 * MrActionBarPopulatorerer is a plugin to fill your action bars, to make it appear that your characters are played regularly by humans.
 * The idea is to add the spells and abilities that a human might use for that particular Class and Specialization.

 * The folder name is important, because it searches for a folder named 'MrActionBarPopulatorerer' that contains XML files, anywhere in your Plugins directory.
     The XML file's filename is not important.
     (eg,  Plugins/MrActionBarPopulatorerer/ActionBar - Mage - Frost.xml  and  Plugins/Downloaded/MrActionBarPopulatorerer/Example Filename.xml  are both acceptable examples)

 * The XML file format is as follows:


<ActionBar>
	<Class>Mage</Class><!-- Required. Character's class. must match Styx.WoWClass - see http://docs.honorbuddy.com/html/49ab5dbc-599e-fc4b-66d7-3a97fd8a543c.htm -->
	<Specialization>MageFrost</Specialization><!-- Required. Character's specialization. must match Styx.WoWSpec - see http://docs.honorbuddy.com/html/e49bf9ec-675c-d6c8-a0f2-5426b27d35fd.htm -->
	<Replace>False</Replace><!-- Optional. Weather or not to replace existing spells in this slot, you probably don't want this. Default: False -->
    <Random></Random><!-- Optional: Weather to place spells randomly ("Random") or semi-randomly within each 12 slot bar ("Group"). Default: None -->
                     <!--   when using random replace is assumed to be False -->
                     <!--   eg: <Random>Random</Random> - If the spell doesn't exist anywhere on any bar, then place it somewhere randomly. Eg, <Slot18>12345</Slot18> the 18 in this case is ignored completely. I highly recommend against this option, as many action bars are invisible and may now have spells in them that dont make sence -->
                     <!--   eg: <Random>Group</Random> - If the spell doesn't exist anywhere on any bar, then place it somewhere randomly within the 12 slot group that it was originally set to be placed in. Eg, <Slot18>12345</Slot18> the spell 12345 will be placed somewhere in the bars 13 to 24 -->
	<Slots>
		<!-- slots are from 1 to 120. all are optional. -->
		<!-- slots 1-12 generally refer to the main action buttons, 61-72 are generally the left bottom action bar, 49-60 the bottom right action bar, 25-36 the top left action bar and 37-48 are the top right action bar -->
        <!-- the other slots usually refer to other specializations or other stances (eg, DruidFeral form)
		<Slot1>spell id</Slot1><!-- optional comment for reference -->
		<Slot2>eg: 84714</Slot2><!-- Frozen Orb -->
		<Slot3>116</Slot3><!-- Frostbolt -->
		...        
        <Slot119>110959</Slot119><!-- Greater Invisibility -->
		<Slot120>113724</Slot120><!-- Ring of Frost -->

	</Slots>
</ActionBar>


 * for testing, you can clear all your action bar using: /run for i = 1,120 do PickupAction(i) ClearCursor() end

*/
//!CompilerOption:Optimize:On
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;

using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.WoWInternals;

using System.Windows.Media;
//requires reference to PresentationCore


namespace MrActionBarPopulatorerer
{
    public class MrActionBarPopulatorerer : Styx.Plugins.HBPlugin
    {
        public override string Name { get { return "MrActionBarPopulatorerer"; } }
        public override string Author { get { return "Any"; } }
        public override Version Version { get { return version; } }
        private static readonly Version version = new Version(0, 4);


        private string Pluginpath = null;
        private int CheckDone = 0;

        

        private void start(EventArgs args)
        {
            CheckDone = 0;
        }


        public enum ActionBarProfilePlacementType { None, Random, Group };
        private class ActionBarProfile
        {
            public WoWClass Class;
            public WoWSpec Spec;

            public bool Replace = false;
            public ActionBarProfilePlacementType Random = ActionBarProfilePlacementType.None;

            public Dictionary<uint, int> Spells = new Dictionary<uint, int>();

            public ActionBarProfile()
            {
                Class = StyxWoW.Me.Class;
                Spec = StyxWoW.Me.Specialization;
            }

            public int PlaceSpellsInActionBar()
            {
                int changes_made = 0;

                Dictionary<uint, int> ActionBarSpells = GetActionBarSpells();
                Random rng = new Random();

                for (uint slot = 1; slot <= 120; slot++)
                {
                    if (Spells.ContainsKey(slot))
                    {
                        int spell = Spells[slot];
                        
                        if (
                            SpellManager.HasSpell(spell) ||
                            Styx.CommonBot.Mount.MountWrapper.FromSpellId(spell) != null
                        )
                        {
                            //ugly code is ugly
                            if (Random == ActionBarProfilePlacementType.Random || Random == ActionBarProfilePlacementType.Group)
                            {
                                if (!ActionBarSpells.Any(s => s.Value == spell))
                                {
                                    bool success = false;
                                    int tries = 0;
                                    do
                                    {
                                        uint newslot = 0;
                                        if (Random == ActionBarProfilePlacementType.Group)
                                        {
                                            double group = Math.Floor((double)(slot-1) / 12) * 12 + 1;//1-12 or 13-24 or 25-36 etc
                                            newslot = (uint)rng.Next((int)group, (int)group + 12);
                                        }
                                        else
                                        {
                                            newslot = (uint)rng.Next(1, 121);
                                        }

                                        if (!ActionBarSpells.ContainsKey(newslot))
                                        {
                                            PlaceSpell(newslot, spell);

                                            if (ActionBarSpells.ContainsKey(newslot))
                                            {
                                                ActionBarSpells.Remove(newslot);
                                            }
                                            ActionBarSpells.Add(newslot, spell);

                                            changes_made++;
                                            success = true;
                                        }
                                        tries++;
                                    }
                                    while (!success && tries < 200);//just an magical arbitary limit, i was worried what might happen if someone's actionbars are completely full
                                    if (!success)
                                    {
                                        Debug("Couldnt find random slot available for spell {0}", spell);
                                    }
                                }
                            }
                            else
                            {
                                if (!ActionBarSpells.ContainsKey(slot))
                                {
                                    PlaceSpell(slot, spell);

                                    if (ActionBarSpells.ContainsKey(slot))
                                    {
                                        ActionBarSpells.Remove(slot);
                                    }
                                    ActionBarSpells.Add(slot, spell);

                                    changes_made++;
                                }
                                else if (Replace && ActionBarSpells[slot] != spell)
                                {
                                    PlaceSpell(slot, spell);

                                    if (ActionBarSpells.ContainsKey(slot))
                                    {
                                        ActionBarSpells.Remove(slot);
                                    }
                                    ActionBarSpells.Add(slot, spell);

                                    changes_made++;
                                }
                            }
                        }
                        else
                        {
                            LogInfo("We dont have the spell {0}", spell);
                        }
                    }
                }

                return changes_made;
            }
        }

        private bool FirstPulse = true;
        public override void Pulse()
        {
            if (FirstPulse)
            {
                FirstPulse = false;

                Pluginpath = FindPluginPath();

                BotEvents.OnBotStarted -= start;
                BotEvents.OnBotStopped -= stop;
                BotEvents.OnBotStarted += start;
                BotEvents.OnBotStopped += stop;
        
                CheckDone = 0;
            }

            if (StyxWoW.Me.IsValid && !StyxWoW.Me.Combat && CheckDone == 0 || CheckDone != StyxWoW.Me.Level)
            {
                CheckDone = StyxWoW.Me.Level;

                

                bool found = false;
                
                //look for the right file for our spec
                foreach (string file in Directory.GetFiles(Pluginpath).Where(f => f.EndsWith(".xml")))
                {
                    //LogInfo("Loading file {0}", file.Replace(Utilities.AssemblyDirectory, ""));
                    try
                    {
                        XDocument doc = XDocument.Load(file);
                        XElement xml = doc.Descendants("ActionBar").FirstOrDefault();

                        if (xml != null && xml.Element("Class") != null && (string)xml.Element("Class") == StyxWoW.Me.Class.ToString() && xml.Element("Specialization") != null && (string)xml.Element("Specialization") == StyxWoW.Me.Specialization.ToString() && xml.Element("Slots") != null)
                        {
                            Log("Loading ActionBar for {0} from {1}", StyxWoW.Me.SafeName, file.Replace(Utilities.AssemblyDirectory, ""));
                            found = true;

                            ActionBarProfile abp = new ActionBarProfile();

                            if (xml.Element("Replace") != null && (string)xml.Element("Replace").Value == "True")
                            {
                                abp.Replace = true;
                            }

                            abp.Random = ActionBarProfilePlacementType.None;
                            if (xml.Element("Random") != null)
                            {
                                if ((string)xml.Element("Random") == "Random")
                                {
                                    abp.Random = ActionBarProfilePlacementType.Random;
                                }
                                else if ((string)xml.Element("Random") == "Group")
                                {
                                    abp.Random = ActionBarProfilePlacementType.Group;
                                }
                                
                            }

                            XElement slots = xml.Element("Slots");
                            
                            for (uint slot = 1; slot <= 120; slot++)
                            {
                                if (slots.Element("Slot"+slot.ToString()) != null)
                                {
                                    int spell;
                                    if (int.TryParse((string)slots.Element("Slot" + slot.ToString()).Value, out spell))
                                    {
                                        abp.Spells.Add(slot,spell);
                                    }
                                    else
                                    {
                                        Err("Error parsing ActionBar slot {0} spell id", slot);
                                    }
                                }
                            }

                            int changes_made = abp.PlaceSpellsInActionBar();
                            if (changes_made > 0)
                            {
                                Log("Added {0} spells to your Action Bars", changes_made);
                            }
                            break;
                        }

                    }
                    catch(Exception e)
                    {
                        Err("Couldn't parse xml file {0} - {1}", file.Replace(Utilities.AssemblyDirectory, ""), e.ToString());
                    }
                }


                if (!found)
                {
                    Log("Creating a dynamic profile for {0} / {1}. Placing spellbook contents into 1-12, 61-72 and 49-60, flying mounts into 25-36 and ground mounts into 37-48.", StyxWoW.Me.Class, StyxWoW.Me.Specialization);

                    ActionBarProfile abp = new ActionBarProfile();
                    abp.Random = ActionBarProfilePlacementType.Group;

                    uint slot = 1;
                    foreach (KeyValuePair<string, WoWSpell> entry in SpellManager.Spells.Where(e => 
                        e.Value.CanCast && 
                        !e.Value.Attributes.HasFlag(SpellAttributes.Passive) &&
                        !e.Value.Name.Contains("Auto")//skip auto attack abilities
                     )
                     .OrderBy(e => e.Value.Attributes.HasFlag(SpellAttributes.CantUsedInCombat))
                     .ThenBy(e => e.Value.Attributes.HasFlag(SpellAttributes.CastableWhileMounted))
                     .ThenBy(e => e.Value.Attributes.HasFlag(SpellAttributes.TradeSkillSpell))
                     .ThenBy(e => e.Value.Id)                     
                     )
                    {
                        abp.Spells.Add(slot,entry.Value.Id);

                        slot++;
                        if (slot == 13)
                        {
                            slot = 61;
                        }
                        else if(slot == 73)
                        {
                            slot = 49;
                        }
                        else if (slot == 61)
                        {
                            slot = 25;
                            break;
                        }
                    }

                    abp.Spells.Add(slot, 150544);
                    slot++;

                    foreach(Styx.CommonBot.Mount.MountWrapper mount in Styx.CommonBot.Mount.FlyingMounts.OrderBy(x => Guid.NewGuid()))
                    {
                        abp.Spells.Add(slot, mount.CreatureSpellId);

                        slot++;
                        if(slot == 37)
                        {
                            break;
                        }
                    }

                    foreach (Styx.CommonBot.Mount.MountWrapper mount in Styx.CommonBot.Mount.GroundMounts)
                    {
                        abp.Spells.Add(slot, mount.CreatureSpellId);

                        slot++;
                        if (slot == 49)
                        {
                            break;
                        }
                    }


                    int NumFavorites = Lua.GetReturnVal<int>(@"local numfav = 0; for i=1,C_MountJournal.GetNumMounts() do local n,id,_,_,u,_,f,_,_,_,c = C_MountJournal.GetMountInfo(i); if c == true then if f == true then numfav = numfav + 1 end end end return(numfav);",0);
                    if (NumFavorites == 0)
                    {
                        LogInfo("You have {0} mounts set as favorite. Setting all of your mounts!", NumFavorites);
                        //for testing//Lua.DoString(@"local numfav = 0; for i=1,C_MountJournal.GetNumMounts() do local n,id,_,_,u,_,f,_,_,_,c = C_MountJournal.GetMountInfo(i); if c == true then if string.match(n, ""Yellow Qiraji"") then print(n); C_MountJournal.SetIsFavorite(i, true) end end end;");
                        Lua.DoString(@"local numfav = 0; for i=1,C_MountJournal.GetNumMounts() do local n,id,_,_,u,_,f,_,_,_,c = C_MountJournal.GetMountInfo(i); if c == true then C_MountJournal.SetIsFavorite(i, true) end end;");
                    }


                    int changes_made = abp.PlaceSpellsInActionBar();
                    if (changes_made > 0)
                    {
                        Log("Added {0} spells to your Action Bars", changes_made);
                    }
                }
            }
        }

        private static void PlaceSpell(uint slot, int spell)
        {
            try
            {
                WoWSpell wowspell = WoWSpell.FromId(spell);//will probably fail for mounts. its just for display purposes anyway
                if (wowspell != null)
                {
                    LogInfo("Placing spell {0}({1}) into slot {2}", wowspell.Name, spell, slot);
                }
                else
                {
                    LogInfo("Placing spell {0} into slot {1}", spell, slot);
                }
            }
            catch (Exception e)
            {
                //ignore!
            }

            Lua.DoString(string.Format("ClearCursor(); PickupSpell({0}); PlaceAction({1}); ClearCursor();", spell, slot));
        }


        private static Dictionary<uint, int> GetActionBarSpells()
        {
            Dictionary<uint, int> ActionBarSpells = new Dictionary<uint, int>();

            Debug("Loading Spells:");
            ActionBarSpells.Clear();

            List<string> spells = Lua.GetReturnValues("local t = {}; for i=1,120 do local type,id,subtype = GetActionInfo(i); if id then local name,_=GetSpellInfo(id); if name then table.insert(t, i..'='..id..'='..name) end end end return unpack(t);");
            if (spells != null)
            {
                foreach (string spell in spells)
                {
                    string[] cols = spell.Split(new string[] { "=" }, StringSplitOptions.None);
                    uint slot;
                    int id;
                    if (uint.TryParse(cols[0], out slot) && int.TryParse(cols[1], out id) && !ActionBarSpells.ContainsKey(slot))
                    {
                        ActionBarSpells.Add(slot, id);
                        Debug("     <Slot{0}>{1}</Slot{0}><!-- {2} -->", slot, id, cols[2]);
                    }
                }
            }

            return ActionBarSpells;
        }

        private void stop(EventArgs args)
        {
            CheckDone = 0;
        }


        private string FindPluginPath(string directory = null)
        {
            if (directory == null)
            {
                directory = Utilities.AssemblyDirectory + @"\Plugins";
            }

            foreach (string subdirectory in Directory.GetDirectories(directory))//couldnt use a lambda expression for some reason?
            {
                if (subdirectory.Contains("MrActionBarPopulatorerer") && Directory.GetFiles(subdirectory).Any(f => f.EndsWith(".xml")))
                {
                    return subdirectory;
                }
                else
                {
                    string temp = FindPluginPath(subdirectory);
                    if (temp != null)
                    {
                        return temp;
                    }
                }
            }

            return null;
        }



        #region logging
        public static void Log(string format, params object[] args)
        {
            Logging.Write(Colors.DodgerBlue, String.Format("MrActionBarPopulatorerer[{0}]: {1}", version, format), args);
            TreeRoot.StatusText = String.Format(format, args);//its bugged. do it twice
            TreeRoot.StatusText = String.Format(format + " ", args);
        }
        public static void LogInfo(string format, params object[] args)
        {
            Logging.Write(Colors.DarkGray, String.Format("MrActionBarPopulatorerer[{0}]: {1}", version, format), args);
        }

        public static void Err(string format, params object[] args)
        {
            Logging.Write(Colors.Red, String.Format("MrActionBarPopulatorerer[{0}]: {1}", version, format), args);
        }

        public static void Debug(string format, params object[] args)
        {
            Logging.WriteDiagnostic(String.Format("MrActionBarPopulatorerer[{0}]: {1}", version, format), args);
        }
        #endregion

       
    }
}
